var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import * as StompJs from '@stomp/stompjs';
import { replaceProtocolInUrl } from './tools';
if (typeof TextEncoder !== 'function') {
    const TextEncodingPolyfill = require('text-encoding');
    global.TextEncoder = TextEncodingPolyfill.TextEncoder;
    global.TextDecoder = TextEncodingPolyfill.TextDecoder;
}
/**
 * Create Stomp on websocket connection to Ordering Stack listening for new orders to process by handlers/callback (onMessageAsync)
 */
export function connectWebSockets(params) {
    return __awaiter(this, void 0, void 0, function* () {
        const { appInsights, debugWs } = params.extra || {};
        let client;
        const baseURL = replaceProtocolInUrl(params.baseURL, 'wss://') + '/websocket';
        let userUUID = '';
        function awaitAccessToken() {
            return __awaiter(this, void 0, void 0, function* () {
                let token = '';
                let tries = 0;
                while (!token) {
                    // ordering-core returns token="" when no token available!
                    const data = yield params.authDataProvider();
                    token = data.token;
                    userUUID = data.UUID;
                    if (!token) {
                        tries++;
                        console.warn('WS no token. Exponential backoff');
                        appInsights === null || appInsights === void 0 ? void 0 : appInsights.trackTrace({
                            message: 'websocket no token. Exponential backoff',
                        }, { tries, venue: params.venue });
                        // exponential backoff
                        yield new Promise((resolve) => setTimeout(resolve, 1000 * Math.min(Math.pow(2, tries), 30)));
                    }
                }
                return token;
            });
        }
        const stompConfig = {
            brokerURL: baseURL,
            debug: function (a) {
                //console.log('DEBUG: ' + a);
            },
            reconnectDelay: 20000,
            heartbeatIncoming: 4000,
            heartbeatOutgoing: 4000,
            beforeConnect: function () {
                return __awaiter(this, void 0, void 0, function* () {
                    console.log('--- beforeConnect --- ');
                    debugWs &&
                        (appInsights === null || appInsights === void 0 ? void 0 : appInsights.trackTrace({ message: 'websocket beforeConnect' }));
                    const token = yield awaitAccessToken();
                    stompConfig.connectHeaders = { login: token /*passcode: ''*/ };
                });
            },
            onConnect: function () {
                return __awaiter(this, void 0, void 0, function* () {
                    if (!stompConfig.connectHeaders) {
                        console.warn('No connect headers');
                        return;
                    }
                    const accessToken = stompConfig.connectHeaders.login;
                    if (params.onKDSMessageAsync) {
                        var subscription = client.subscribe(`/kds/${params.tenant}/${params.venue}`, function (data) {
                            return __awaiter(this, void 0, void 0, function* () {
                                var message = JSON.parse(data.body);
                                yield params.onKDSMessageAsync(message);
                            });
                        });
                    }
                    if (params.onOrdersUpdateAsync) {
                        var subscriptionForOrdersUpdate = client.subscribe(`/order-changes/${params.tenant}/${userUUID}`, function (data) {
                            return __awaiter(this, void 0, void 0, function* () {
                                var message = JSON.parse(data.body);
                                yield params.onOrdersUpdateAsync(message);
                            });
                        }, params.venue ? { 'x-venue': params.venue } : undefined);
                    }
                    if (params.onNotificationAsync) {
                        var subscriptionForNotifications = client.subscribe(`/notifications/${params.tenant}/${userUUID}`, function (data) {
                            return __awaiter(this, void 0, void 0, function* () {
                                var message = JSON.parse(data.body);
                                yield params.onNotificationAsync(message);
                            });
                        }, params.venue ? { 'x-venue': params.venue } : undefined);
                    }
                    if (params.onSteeringCommand && params.venue) {
                        client.subscribe(`/steering/${params.tenant}/${params.venue}`, function (data) {
                            return __awaiter(this, void 0, void 0, function* () {
                                var _a;
                                var message = JSON.parse(data.body);
                                (_a = params.onSteeringCommand) === null || _a === void 0 ? void 0 : _a.call(params, message);
                            });
                        });
                    }
                    console.log('Websocket connected.');
                    // notify connected only after all subscriptions are set up
                    yield params.onConnectedAsync(accessToken);
                });
            },
            onDisconnect: function () {
                return __awaiter(this, void 0, void 0, function* () {
                    yield params.onDisconnectAsync();
                    console.log('Websocket disconnected.');
                });
            },
            onStompError: function (frame) {
                return __awaiter(this, void 0, void 0, function* () {
                    // Will be invoked in case of error encountered at Broker
                    // Bad login/passcode typically will cause an error
                    // Complaint brokers will set `message` header with a brief message. Body may contain details.
                    // Compliant brokers will terminate the connection after any error
                    console.error('onStompError headers.message: ' + (frame === null || frame === void 0 ? void 0 : frame.headers['message']));
                    debugWs &&
                        (appInsights === null || appInsights === void 0 ? void 0 : appInsights.trackTrace({
                            message: 'websocket onStompError headers.message: ' +
                                (frame === null || frame === void 0 ? void 0 : frame.headers['message']),
                        }));
                    console.error('onStompError body: ' + frame.body);
                    yield client.deactivate();
                    client = createNewClient(stompConfig);
                    client.activate();
                });
            },
            onWebSocketClose: function () {
                debugWs &&
                    (appInsights === null || appInsights === void 0 ? void 0 : appInsights.trackTrace({
                        message: 'websocket closed.',
                    }));
                console.log('Websocket closed.');
            },
            onWebSocketError: function (error) {
                debugWs &&
                    (appInsights === null || appInsights === void 0 ? void 0 : appInsights.trackTrace({
                        message: 'websocket error',
                        properties: { error },
                    }));
                console.log('Websocket error.');
            },
            onUnhandledMessage: function (message) {
                debugWs &&
                    (appInsights === null || appInsights === void 0 ? void 0 : appInsights.trackTrace({
                        message: 'websocket onUnhandledMessage',
                        properties: { message },
                    }));
            },
            logRawCommunication: true,
            discardWebsocketOnCommFailure: true,
        };
        try {
            const token = yield awaitAccessToken();
            stompConfig.connectHeaders = { login: token /*passcode: ''*/ };
            client = createNewClient(stompConfig);
            client.activate();
            return () => __awaiter(this, void 0, void 0, function* () {
                yield client.deactivate();
                console.log('Websocket connection deactivated');
            });
        }
        catch (err) {
            console.error(err);
            return () => Promise.resolve();
        }
    });
}
function createNewClient(stompConfig) {
    const _client = new StompJs.Client(stompConfig);
    if (typeof WebSocket !== 'function') {
        // Fallback code
        _client.webSocketFactory = () => {
            const WebSocket = require('ws');
            global.WebSocket = WebSocket;
            //console.log('create ws with ' + stompConfig.brokerURL);
            const ws = new WebSocket(stompConfig.brokerURL);
            return ws;
        };
    }
    return _client;
}
